/*----------------------------------------------------------------------------

  Point alignment detector (3D)

  This code is part of the following IPOL publication and was subject
  to peer review. If you use it, please cite it as:

   .........................................

  Copyright (c) 2017  Alvaro Gomez  <agomez@fing.edu.uy>,
  Gregory Randall<randall@fing.edu.uy> and Rafael Grompone von Gioi <grompone@gmail.com>.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>.


  ----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "util.h"


//------------------------------------------------------------------------------
// Use of sectors to divide the local cylinder - global variables
int numberOfSectors;        // number of sectors
double sectorAngle;         // angle of a sector = (2*M_PI/numberOfSectors);
//-------------------------


/*----------------------------------------------------------------------------*/
/** max and min value */
#define max(a,b) (((a)>(b))?(a):(b))
#define min(a,b) (((a)<(b))?(a):(b))


/*----------------------------------------------------------------------------*/
/** Fatal error, print a message to standard-error output and exit.
 */
static void error(char * msg)
{
	fprintf(stderr,"Error: %s\n",msg);
	exit(EXIT_FAILURE);
}

/*----------------------------------------------------------------------------*/
/** Memory allocation, print an error and exit if fail.
 */
static void * xmalloc(size_t size)
{
	void * p;
	if( size == 0 ) error("xmalloc: zero size");
	p = malloc(size);
	if( p == NULL ) error("xmalloc: out of memory");
	return p;
}

/*----------------------------------------------------------------------------*/
/** Memory re-allocation, print an error and exit if fail.
 */
static void * xrealloc(void * p, size_t size)
{
	if( size == 0 ) error("xrealloc: zero size");
	p = realloc(p,size);
	if( p == NULL ) error("xrealloc: out of memory");
	return p;
}

/*----------------------------------------------------------------------------*/
/* Log10 of tail of binomial distribution by Hoeffding approximation
 */
static double log_bin(int n, int k, double p)
{
	double r = (double) k / (double) n;
	if( r <= p ) return 0.0;
	else if( n == k) return k * log10(p);
	else return k * log10(p/r) + (n-k) * log10( (1-p)/(1-r) );
}

/*----------------------------------------------------------------------------*/
static int comp_align(const void *a, const void *b)
{
	if( *( ((double *) a) + 9) < *( ((double *) b) + 9) ) return -1;
	if( *( ((double *) a) + 9) > *( ((double *) b) + 9) ) return  1;
	return 0;
}

/*----------------------------------------------------------------------------*/
static int maxInArray(const int *a, const int length)
{
	int max = -1;
	int i=0;
	for(i=0;i<length;i++){
		if (a[i]>max)
			max=a[i];
	}
	return max;
}
/*----------------------------------------------------------------------------*/
static void zeroArray(int *a, const int length)
{
	int i=0;
	for(i=0;i<length;i++){
		a[i]=0;
	}
}


/*----------------------------------------------------------------------------*/
// Checks if a point (given by its index m) is part of an alignment or the locallity
// of the alignment
// Returns 0 if the point has nothing to do with the alignment. Otherwise returns 1 if
// the point is in the alignment cylinder or 2 if the point is in the local cylinder
static int check_point_wrt_alignment(double * point, int N,     //the N points
									  int m,                     //index of test point
									  double x1, double y1, double z1, //origin of alignment
									  double dx, double dy, double dz, //unit vector of alignment
									  double len,                    //length of alignment
									  double orthA[],double orthB[], //ref frame of alignment
									  double w, double l,      //tested diameters
									  int sectorAccumulator[], //sector accumulator array
									  double* in_align, int *p_nn, double *p_xx)   //in_align array and counter
{
	double x0, y0, z0, xx, yy, angle, aux, auxi;
	int sector;

	/* compute coordinates of the point p0 relative to the alignment p1p2*/
	x0=point[3*m]; y0=point[3*m+1]; z0=point[3*m+2];
	xx =  dx * (x0-x1) + dy * (y0-y1) + dz * (z0-z1) ; //dot product for the projection on the alignment p1p2
	//save the xx value
	(*p_xx)=xx;
	//cross product for the distance to p1p2 [dy*(z0-z1) - dz*(y0-y1), -dx*(z0-z1)+dz*(x0-x1), dx*(y0-y1)-dy*(x0-x1)]
	yy=0;
	aux =   dy*(z0-z1) - dz*(y0-y1); yy+= aux*aux;
	aux =  -dx*(z0-z1) + dz*(x0-x1); yy+= aux*aux;
	aux =   dx*(y0-y1) - dy*(x0-x1); yy+= aux*aux;
	yy= sqrt(yy);

	/* check if the point is in the local cylinder */
	if( xx < 0.0 || xx >= len )
		return(0);
	if( yy > l/2.0 )
		return(0);


	if( yy > w/2.0 ){
		// if the point is in the local cylinder and out of the alignment cylinder

		// compute the angle in the plane (orthA,orthB)
		aux = orthA[0] * (x0-x1) + orthA[1] * (y0-y1) + orthB[2] * (z0-z1) ; //dot product with orthA
		auxi= orthB[0] * (x0-x1) + orthB[1] * (y0-y1) + orthB[2] * (z0-z1) ; //dot product with orthB
		angle = atan2(auxi,aux); // range -pi/2, pi/2
		angle = (angle>=0 ? angle : 2*M_PI + angle); // range 0, 2*pi

		// compute the sector index (integer division)
		sector = angle/sectorAngle;
		//increment the acccumulator
		sectorAccumulator[sector]++;
		return(2);
	}
	else {
		// if the point is in the alignment cylinder
		//and if it is a real point, not an extended point (extended point are in m>=N),
		// add to in_align the distance on the alignment (it will be used later to
		// check in what pillbox is the point)
		if(m<N){
			if (in_align!=0)
				in_align[(*p_nn)] = xx;
			(*p_nn)++;
		}
		return(1);
	}

	//just in case there is an error and we arrive here
	return(-1);

}

/*----------------------------------------------------------------------------*/
/*
   input:  point           - input points (including extended points at the end)
           N               - number of points
           Next            - number of extended points
           X,Y,Z           - domain size [0,X] x [0,Y] x [0,Z]

           (arguments to test a single alignment, if both > 0)
           indexOfPoint1   - index of point1 in the alignment
           indexOfPoint2   - index of point2 in the alignment

   output: pointer to *Nout 10-tuples
           x1 y1 z1 x2 y2 z2 alignment_cylinder_diameter local_cylinder_diameter num_boxes log_nfa

 */
static double * find_alignments( double * point, int N, int Next, int * Nout,
		double X, double Y, double Z ,
		int indexOfPoint1, int indexOfPoint2)
{



	int i,j,k,n,nn,m,v,b,n_box;
	double x1,y1,x2,y2,z1,z2,dx,dy,dz,len,w,xx,yy,logNFA,l,box,p0,p1,d;
	double max_x, max_y, max_z,min_x, min_y, min_z;
	double x0,y0,z0,aux,auxi;
	double max_w_over_len = 10.0;
	int n_w = 8;
	int n_l = 8;
	int ww,ll;
	int NN;
	int * used;
	double * in_align;
	int ret;

	//----sectors---------------------
	int sectorAccumulator[360];     // pointer to the accumulator (count of points in each sector)
	int sector;                 // variable holding a sector index
	double orthA[3], orthB[3];  // vectors orthogonal to the alignment
	double angle;               // variable holding an angle
	//--------------------------------

	/* number of tests */
	double logNT = log10(N) + log10(N-1) - log10(2) + log10(n_w) + log10(n_l)
                				 + log10(N)/3.0;

	/* initialize output */
	int n_align = 0;
	int n_alloc = 1;
	double * align = xmalloc( n_alloc * 10 * sizeof(double) );

#pragma omp parallel for schedule(dynamic) private(i,j,x1,y1,x2,y2,z1,z2,dx,dy,dz,len,w,ww,l,ll,d,NN,nn,m,xx,yy,in_align,k,box,v,used,b,n,p0,p1,n_box,logNFA, x0,y0,z0,aux,auxi,max_x, max_y, max_z,min_x, min_y, min_z, sectorAccumulator, sector,orthA, orthB, angle, ret )
	/* loop over all triplets of points - Algorithm 2 - lines 1,2 */
	for(i=0; i<(N-1); i++)
		for(j=i+1; j<N; j++)
		{
			// if a test of a single alignment, consider only that pair of points.
			if (indexOfPoint1>=0 && (indexOfPoint1 !=i || indexOfPoint2 !=j) )
				continue;

			x1 = point[3*i];          y1 = point[3*i+1];          z1 = point[3*i+2];
			x2 = point[3*j];          y2 = point[3*j+1];          z2 = point[3*j+2];
			dx = x2-x1;               dy = y2-y1;               dz = z2-z1;
			len = sqrt(dx*dx+dy*dy+dz*dz);
			if( len == 0.0 )  continue;
			dx /= len;                dy /= len;                dz /= len;

			//Compute the reference frame for the alignment (orthA,orthB,(dx,dy,dz))
			//get orthogonal direction to (dx,dy,dz)
			// (dz,dz,-dx-dy) or (-dy-dz,dx,dx)
			aux = sqrt(2*dz*dz + (dx+dy)*(dx+dy)); //norm of (dz,dz,-dx-dy)
			auxi = sqrt(2*dx*dx + (dy+dz)*(dy+dz)); //norm of (-dy-dz,dx,dx)
			if (aux > auxi){
				orthA[0]=dz/aux; orthA[1]=dz/aux; orthA[2]=(-dx-dy)/aux;
			}
			else{
				orthA[0]=(-dy-dz)/auxi; orthA[1]=dx/auxi; orthA[2]=dx/auxi;
			}
			//orthB = cross((dx,dy,dz), orthA)
			orthB[0] =  dy*orthA[2] - dz*orthA[1];
			orthB[1] = -dx*orthA[2] + dz*orthA[0];
			orthB[2] =  dx*orthA[1] - dy*orthA[0];


			/* get memory */
			used = xmalloc( (N-1) * sizeof(int) );
			in_align = xmalloc( N * sizeof(double) );

			/* loop over width - Algorithm 2 - line 5*/
			for( w=len/max_w_over_len, ww=0; ww<n_w; ++ww, w/=sqrt(sqrt(2)) )
			{
				/* loop over local window size - Algorithm 2 - line 8*/
				for( l=len/sqrt(10), ll=0; l>w && ll<n_l; ++ll, l/=sqrt(sqrt(2)) )
				{

					//if NOT FLANN use the Next points if the cylinder is out of the bounds
					max_x=max(x1,x2)+l/2.0; max_y=max(y1,y2)+l/2.0; max_z=max(z1,z2)+l/2.0;
					min_x=min(x1,x2)-l/2.0; min_y=min(y1,y2)-l/2.0; min_z=min(z1,z2)-l/2.0;
					if (
							min_x < 0.0 || max_x > X ||
							min_y < 0.0 || max_y > Y ||
							min_z < 0.0 || max_z > Z 
					)
						NN = Next;
					else
						NN = N;

					/* initialize the count of points in an alignment */
					nn = 0;

					// initialize the sector accumulator
					zeroArray(sectorAccumulator,numberOfSectors);

					//if FLANN is NOT in use, use all the points
					for(m=0; m<NN; m++)
					{
						if( m==i || m==j ) continue; /* do not count the points that define the alignment*/

						ret = check_point_wrt_alignment(  point, N,           //the N points
								m,                  //index of test point
								x1, y1, z1,         //origin of alignment
								dx, dy,  dz,        //unit vector of alignment
								len,                //length of alignment
								orthA, orthB,       //ref frame of alignment
								w,  l,              //tested diameters
								sectorAccumulator,  //sector accumulator array
								in_align, &nn, &xx);
					}

					/* number of boxes in the alignment: from 1 to N-2
					the two points that define the alignment are not counted
					Algorithm 2 - For loop - lines 12-21
					*/
					for(k=max(nn/2,1); k<=min(2*nn,N-2); k++)
					{
						/* compute box side */
						box = len / (k+1);

						/* initialize used boxes */
						for(v=0; v<k; v++) used[v] = 0;

						/* count points in boxes */
						for(m=0; m<nn; m++)
						{
							b = floor( (in_align[m] - box/2.0) / box );
							if( b>=0 && b<k ) ++used[b];
						}

						//conservative estimation of the number of points in the local cylinder
						n = numberOfSectors * maxInArray(sectorAccumulator, numberOfSectors) + nn;

						p0 = (box*w*w) / (len*l*l);
						p1 = 1.0 - pow( 1.0 - p0, n );

						for( n_box = 0, v=0; v<k; v++ )
							if( used[v] != 0 )
								++n_box;
						logNFA = logNT + log_bin( k, n_box, p1 );

						/* store result */
						if( logNFA < 0.0 )
#pragma omp critical
						{
							/* get more memory if needed */
							if( n_align >= n_alloc )
							{
								n_alloc *= 2;
								align = xrealloc(align, n_alloc*10*sizeof(double));
							}

							align[10*n_align+0] = x1;
							align[10*n_align+1] = y1;
							align[10*n_align+2] = z1;
							align[10*n_align+3] = x2;
							align[10*n_align+4] = y2;
							align[10*n_align+5] = z2;
							align[10*n_align+6] = w;
							align[10*n_align+7] = l;
							align[10*n_align+8] = k;
							align[10*n_align+9] = logNFA;
							++n_align;
						}
					}
				}
			}

			/* free memory */
			free( (void *) used );
			free( (void *) in_align );
		}

	/* result */
	*Nout = n_align;
	return align;
}

/*----------------------------------------------------------------------------*/
/*
input:  point     - input points (including extended points at the end)
	   N         - number of points
	   Next      - number of extended points
	   align     - list of alignments
	   Na        - number of input alignments
	   X,Y,Z     - domain size [0,X] x [0,Y] x [0,Z]

output: pointer to *Nout 10-tuples
	   x1 y1 z1 x2 y2 z2 alignment_cylinder_diameter local_cylinder_diameter num_boxes log_nfa

 */
static double * masking( double * point, int N, int Next,
		double * align, int Na, int * Nout,
		double X, double Y, double Z )
{
	int i,j,k,n,nn,m,v,b,n_box;
	double x1,y1,x2,y2,dx,dy,len,w,xx,yy,l,box,p0,p1,d,z1,z2,dz, x0,y0,z0, aux, auxi;
	double m_x1,m_y1,m_x2,m_y2,m_dx,m_dy,m_len,m_w,m_xx,m_yy,m_l, m_z1,m_z2, m_dz;
	double max_x, max_y, max_z,min_x, min_y, min_z;
	double logNFA,max_logNFA;
	int n_w = 8;
	int n_l = 8;
	int NN;
	int ret;

	//----sectors---------------------
	int sectorAccumulator[360];     // pointer to the accumulator (count of points in each sector)
	int sectorAccumulatorBis[360];


	int sector;                 // variable holding a sector index
	double orthA[3], orthB[3];  // vectors orthogonal to the alignment
	double angle;               // variable holding an angle
	//--------------------------------

	/* number of tests */
	double logNT = log10(N) + log10(N-1) - log10(2) + log10(n_w) + log10(n_l)
							 + log10(N)/3.0;

	/* get memory to store used boxes */
	int * used = xmalloc( (N-1) * sizeof(int) );

	/* initialize output */
	int n_out = 0;
	int n_alloc = 1;
	double * out = xmalloc( n_alloc * 10 * sizeof(double) );

	/* sort alignments by NFA */
	qsort( (void *) align, Na, 10*sizeof(double), &comp_align );

	/* first alignment is directly added to the output */
	out[10*n_out+0] = align[0];
	out[10*n_out+1] = align[1];
	out[10*n_out+2] = align[2];
	out[10*n_out+3] = align[3];
	out[10*n_out+4] = align[4];
	out[10*n_out+5] = align[5];
	out[10*n_out+6] = align[6];
	out[10*n_out+7] = align[7];
	out[10*n_out+8] = align[8];
	out[10*n_out+9] = align[9];

	++n_out;

	/* loop over ordered alignments */
	for(i=1; i<Na; i++)
	{

		/* alignment data */
		x1 = align[10*i+0];
		y1 = align[10*i+1];
		z1 = align[10*i+2];
		x2 = align[10*i+3];
		y2 = align[10*i+4];
		z2 = align[10*i+5];
		w  = align[10*i+6];
		l  = align[10*i+7];
		k  = align[10*i+8];
		dx = x2-x1;
		dy = y2-y1;
		dz = z2-z1;
		len = sqrt(dx*dx+dy*dy+dz*dz);
		dx /= len;
		dy /= len;
		dz /= len;
		box = len / (k+1);
		max_logNFA = align[10*i+9];


		//Compute the reference frame for the alignment (orthA,orthB,(dx,dy,dz))
		//get orthogonal direction to (dx,dy,dz)
		// (dz,dz,-dx-dy) or (-dy-dz,dx,dx)
		aux = sqrt(2*dz*dz + (dx+dy)*(dx+dy)); //norm
		auxi = sqrt(2*dx*dx + (dy+dz)*(dy+dz)); //norm
		if (aux > auxi){
			orthA[0]=dz/aux; orthA[1]=dz/aux; orthA[2]=(-dx-dy)/aux;
		}
		else{
			orthA[0]=(-dy-dz)/auxi; orthA[1]=dx/auxi; orthA[2]=dx/auxi;
		}
		//orthB = cross((dx,dy,dz), orthA)
		orthB[0] =  dy*orthA[2] - dz*orthA[1];
		orthB[1] = -dx*orthA[2] + dz*orthA[0];
		orthB[2] =  dx*orthA[1] - dy*orthA[0];

		/* try masking the alignment by already detected ones */

		for(j=0; j<n_out; j++)
		{
			/* masking candidate data */
			m_x1 = out[10*j+0];
			m_y1 = out[10*j+1];
			m_z1 = out[10*j+2];
			m_x2 = out[10*j+3];
			m_y2 = out[10*j+4];
			m_z2 = out[10*j+5];

			m_w  = out[10*j+6];
			m_l  = out[10*j+7];
			m_dx = m_x2-m_x1;
			m_dy = m_y2-m_y1;
			m_dz = m_z2-m_z1;
			m_len = sqrt(m_dx*m_dx+m_dy*m_dy+m_dz*m_dz);
			m_dx /= m_len;
			m_dy /= m_len;
			m_dz /= m_len;

			/* initialize used boxes */
			for(v=0; v<k; v++) used[v] = 0;

			
			//if NOT FLANN use the Next points if the cylinder is out of the bounds
			max_x=max(x1,x2)+l/2.0; max_y=max(y1,y2)+l/2.0; max_z=max(z1,z2)+l/2.0;
			min_x=min(x1,x2)-l/2.0; min_y=min(y1,y2)-l/2.0; min_z=min(z1,z2)-l/2.0;
			if (
					min_x < 0.0 || max_x > X ||
					min_y < 0.0 || max_y > Y ||
					min_z < 0.0 || max_z > Z 
			)
				NN = Next;
			else
				NN = N;


			/* initialize the count of points in an alignment */
			nn = 0;
			int nnBis=0;

			//initialize the accumulator
			zeroArray(sectorAccumulator,numberOfSectors);

			for(m=0; m<NN; m++)
			{

				x0=point[3*m]; y0=point[3*m+1]; z0=point[3*m+2];
				/* do not count defining points */
				if( x0 == x1 && y0 == y1 && z0 == z1 ) continue;
				if( x0 == x2 && y0 == y2 && z0 == z2 ) continue;

				ret = check_point_wrt_alignment(point, N,           //the N points
						m,                  //index of test point
						x1, y1, z1,         //origin of alignment
						dx, dy,  dz,        //unit vector of alignment
						len,                //length of alignment
						orthA, orthB,       //ref frame of alignment
						w,  l,              //tested diameters
						sectorAccumulator,  //sector accumulator array
						0, &nn, &xx);       //in_align=0, no align array

				//only consider a point that belongs to the alignment
				if(ret!=1)
					continue;

				// do not count extended points
				if( m>=N ) continue;

				/* compute coordinates of the point p0 relative to the masking alignment candidate m_p1 m_p2*/
				m_xx =  m_dx * (x0-m_x1) + m_dy * (y0-m_y1) +m_dz * (z0-m_z1) ; //dot product;

				//cross product for the distance to the alignment
				m_yy=0;
				aux =   m_dy*(z0-m_z1) - m_dz*(y0-m_y1); m_yy+= aux*aux;
				aux =  -m_dx*(z0-m_z1) + m_dz*(x0-m_x1); m_yy+= aux*aux;
				aux =   m_dx*(y0-m_y1) - m_dy*(x0-m_x1); m_yy+= aux*aux;
				m_yy= sqrt(m_yy);

				// check if the point belongs to the alignment cylinder of the masking candidate
				// in that case do not consider the point
				if( fabs(m_yy) < m_w/2.0 && m_xx>=0.0 && m_xx<m_len ) continue;

				/* compute the box for the point */
				b = floor( (xx - box/2.0) / box );
				if( b>=0 && b<k ) ++used[b];
			}

			/* compute NFA after masking */
			n = numberOfSectors * maxInArray(sectorAccumulator, numberOfSectors) + nn;


			p0 = (box*w*w) / (len*l*l);
			p1 = 1.0 - pow( 1.0 - p0, n );
			for( n_box = 0, v=0; v<k; v++ )
				if( used[v] != 0 )
					++n_box;
			logNFA = logNT + log_bin( k, n_box, p1 );
			if( logNFA > max_logNFA ) max_logNFA = logNFA;
		}
		if( max_logNFA > 0.0 ) continue;

		/* get more memory if needed */
		if( n_out >= n_alloc )
		{
			n_alloc *= 2;
			out = xrealloc( out, n_alloc * 10 * sizeof(double) );
		}

		/* store alignment to output */
		out[10*n_out+0] = align[10*i+0];
		out[10*n_out+1] = align[10*i+1];
		out[10*n_out+2] = align[10*i+2];
		out[10*n_out+3] = align[10*i+3];
		out[10*n_out+4] = align[10*i+4];
		out[10*n_out+5] = align[10*i+5];
		out[10*n_out+6] = align[10*i+6];
		out[10*n_out+7] = align[10*i+7];
		out[10*n_out+8] = align[10*i+8];
		out[10*n_out+9] = align[10*i+9];
		++n_out;
	}

	/* free memory */
	free( (void *) used );

	/* result */
	*Nout = n_out;
	return out;
}

/*----------------------------------------------------------------------------*/
/*
input:  point           - input points
   N               - number of points
   X,Y,Z           - domain size [0,X] x [0,Y] x [0,Z]
   sectors         - number of sectors to divide the local cylinder into
   indexOfPoint1   - (to test a single alignment) index of point1 in the alignment
   indexOfPoint2   - (to test a single alignment) index of point2 in the alignment

output: pointer to *Nout 10-tuples
   x1 y1 z1 x2 y2 z2 width local_window_width num_boxes log_nfa

 */
double * point_alignments(double * point, int N, double X, double Y, double Z, int * Nout, int sectors, int indexOfPoint1, int indexOfPoint2)
{
	double * align;
	double * align_m;
	int Na,Nm;
	double * point_ext;
	int Next;
	int i,j;

	// Set the global variable that holds the number of sectors to divide the local cylinder
	numberOfSectors = sectors;
	sectorAngle = 2*M_PI/numberOfSectors;



	/* Extend the points to handle eventual problems in the domain border:
	 * Order of the 27 extension regions
	 *
	 *  z=0                       z=-1                      z=1
	 * |  1  |  2  |  3  |      |  10 |  11 |  12 |      |  19 |  20 |  21 |
	 * |  4  |  0  |  5  |      |  13 |  9  |  14 |      |  22 |  18 |  23 |
	 * |  6  |  7  |  8  |      |  15 |  16 |  17 |      |  25 |  25 |  26 |
	 *
	 */
	double Xsign[27]=    { 1, -1,  1, -1, -1, -1, -1,  1, -1,    1, -1,  1, -1, -1, -1, -1,  1, -1,    1, -1,  1, -1, -1, -1, -1,  1, -1 };
	double Ysign[27] =   { 1, -1, -1, -1,  1,  1, -1, -1, -1,    1, -1, -1, -1,  1,  1, -1, -1, -1,    1, -1, -1, -1,  1,  1, -1, -1, -1 };
	double Zsign[27] =   { 1,  1,  1,  1,  1,  1,  1,  1,  1,   -1, -1, -1, -1, -1, -1, -1, -1, -1,   -1, -1, -1, -1, -1, -1, -1, -1, -1 };
	double Xoffset[27] = { 0,  0,  0,  2,  0,  2,  0,  0,  2,    0,  0,  0,  2,  0,  2,  0,  0,  2,    0,  0,  0,  2,  0,  2,  0,  0,  2 };
	double Yoffset[27] = { 0,  0,  0,  0,  0,  0,  2,  2,  2,    0,  0,  0,  0,  0,  0,  2,  2,  2,    0,  0,  0,  0,  0,  0,  2,  2,  2 };
	double Zoffset[27] = { 0,  0,  0,  0,  0,  0,  0,  0,  0,    0,  0,  0,  0,  0,  0,  0,  0,  0,    2,  2,  2,  2,  2,  2,  2,  2,  2 };

	/* extended points to handle domain border: symmetric extension */
	Next = 27*N;
	point_ext = xmalloc( 3*Next * sizeof(double) );
	for(i=0; i<27; i++)
		for(j=0; j<N; j++)
		{
			point_ext[i*3*N + 3*j + 0] = Xsign[i] * point[3*j + 0] + X*Xoffset[i];
			point_ext[i*3*N + 3*j + 1] = Ysign[i] * point[3*j + 1] + Y*Yoffset[i];
			point_ext[i*3*N + 3*j + 2] = Zsign[i] * point[3*j + 2] + Z*Zoffset[i];

		}

	printf("First step: find_alignments BEGIN.\n");
	clockStart();
	/* find all meaningful alignments */
	align = find_alignments(point_ext,N,Next,&Na,X,Y,Z, indexOfPoint1, indexOfPoint2);
	clockStop("First step: find_alignments END");
	printf("First step: find_alignments: ALIGNMENTS = %d\n", Na);


	printf("Second step: masking BEGIN.\n");
	clockStart();
	/* reduce redundancy */
	if( Na>=1 ) align_m = masking(point_ext,N,Next,align,Na,&Nm,X,Y,Z);
	clockStop("Second step: masking END");
	printf("Second step: masking ALIGNMENTS = %d\n", Nm);


	/* free memory */
	free( (void *) align );
	free( (void *) point_ext );

	/* result */
	if( Na>=1 && Nm>=1 )
	{
		*Nout = Nm;
		return align_m;
	}

	/* no detection */
	*Nout = 0;
	return NULL;
}
/*----------------------------------------------------------------------------*/
