/*----------------------------------------------------------------------------

  Point alignment detector (3D)

  This code is part of the following IPOL publication and was subject
  to peer review. If you use it, please cite it as:

  ...............................................

  Copyright (c) 2017  Alvaro Gomez  <agomez@fing.edu.uy>,
  Gregory Randall<randall@fing.edu.uy> and Rafael Grompone von Gioi <grompone@gmail.com>.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>.

 ----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include "point_alignments.h"

/*----------------------------------------------------------------------------*/
/** max and min value */
#define max(a,b) (((a)>(b))?(a):(b))
#define min(a,b) (((a)<(b))?(a):(b))

/*----------------------------------------------------------------------------*/
/** Fatal error, print a message to standard-error output and exit.
 */
static void error(char * msg)
{
  fprintf(stderr,"Error: %s\n",msg);
  exit(EXIT_FAILURE);
}

/*----------------------------------------------------------------------------*/
/** Memory allocation, print an error and exit if fail.
 */
static void * xmalloc(size_t size)
{
  void * p;
  if( size == 0 ) error("xmalloc: zero size");
  p = malloc(size);
  if( p == NULL ) error("xmalloc: out of memory");
  return p;
}

/*----------------------------------------------------------------------------*/
/** Memory re-allocation, print an error and exit if fail.
 */
static void * xrealloc(void * p, size_t size)
{
  if( size == 0 ) error("xrealloc: zero size");
  p = realloc(p,size);
  if( p == NULL ) error("xrealloc: out of memory");
  return p;
}

/*----------------------------------------------------------------------------*/
/** Open file, print an error and exit if fail.
 */
static FILE * xfopen(const char * path, const char * mode)
{
  FILE * f = fopen(path,mode);
  if( f == NULL ) error("xfopen: unable to open file");
  return f;
}

/*----------------------------------------------------------------------------*/
/** Close file, print an error and exit if fail.
 */
static int xfclose(FILE * f)
{
  if( fclose(f) == EOF ) error("xfclose: unable to close file");
  return 0;
}

/*----------------------------------------------------------------------------*/
static double * read_txt(char * filename, int * N)
{
  FILE * f;
  int n,n_alloc;
  double val;
  double * p;

  /* open file */
  f = xfopen(filename,"r");

  /* initialize output */
  n = 0;
  n_alloc = 1;
  p = xmalloc( n_alloc * sizeof(double) );

  /* read points */
  while( fscanf(f,"%lf%*[^0-9.eE+-]",&val) == 1 )
    {
      if( n >= n_alloc )
        {
          n_alloc *= 2;
          p = xrealloc( p, n_alloc * sizeof(double) );
        }
      p[n++] = val;
    }

  /* close file */
  xfclose(f);

  /* return list of points */
  *N = n;
  return p;
}

/*----------------------------------------------------------------------------*/
/** Read a TXT file with data in three columns
 *  Discard comment lines starting with #
 */
static double * read_txt_handle_comments(char * filename, int * N)
{
  FILE * f;
  int i,n,n_alloc;
  double val[3];
  double * p;
  char line[1024];
  int valuesInLine;

  /* open file */
  f = xfopen(filename,"r");

  /* initialize output */
  n = 0;
  n_alloc = 1;
  p = xmalloc( n_alloc * sizeof(double) );

  /* read points */
  while( fgets(line, sizeof line, f) )
    {
	  if (line[0]=='#')
		  continue;
	  valuesInLine=sscanf(line,"%lf%*[^0-9.eE+-]%lf%*[^0-9.eE+-]%lf%*[^0-9.eE+-]",&val[0],&val[1],&val[2]);
	  for (i=0; i<valuesInLine; i++)
	  {
		  p[n++] = val[i];
		  if( n >= n_alloc )
		  {
			  n_alloc *= 2;
			  p = xrealloc( p, n_alloc * sizeof(double) );
		  }
	  }

    }

  /* close file */
  xfclose(f);

  /* return list of points */
  *N = n;
  return p;
}


/*----------------------------------------------------------------------------*/
/** Write an TXT format matrix file
 */
void write_txt(double * data, int X, int Y, char * filename)
{
  FILE * f;
  int i,j;

  /* open file */
  f = xfopen(filename,"w");

  /* write data */
  if( data != NULL && X > 0 && Y > 0 )
    for(j=0; j<Y; j++)
      {
        for(i=0; i<X; i++) fprintf(f,"%.16g ",data[i+j*X]);
        fprintf(f,"\n");
      }

  /* close file */
  xfclose(f);
}

/*----------------------------------------------------------------------------*/
/** Get the domain of the points [minX,maxX], [minY,maxY], [minZ,maxZ]
 *  
 *
 */
void get_points_domain(double * point, int numberOfPoints, double *minX, double *maxX, double *minY, double *maxY, double *minZ, double *maxZ){
	int i;
	double x,y,z;
	double dx,dy,dz;
	double marginScale = 0.1; // max=maxCoord + (maxCoord-minCoord)*marginScale; min = minCoord - (maxCoord-minCoord)*marginScale;
	*minX=*minY=*minZ=DBL_MAX;
	*maxX=*maxY=*maxZ=-DBL_MAX;

	for(i=0; i<numberOfPoints; i++){
		x=point[i*3]; y=point[i*3+1]; z=point[i*3+2];
		if (x<*minX) *minX=x;
		if (x>*maxX) *maxX=x;
		if (y<*minY) *minY=y;
		if (y>*maxY) *maxY=y;
		if (z<*minZ) *minZ=z;
		if (z>*maxZ) *maxZ=z;
	}

	//add a margin to the limits
	dx=*maxX-*minX;
	dy=*maxY-*minY;
	dz=*maxZ-*minZ;
	*minX -= dx*marginScale; 
	*minY -= dy*marginScale;
	*minZ -= dz*marginScale;
	*maxX += dx*marginScale; 
	*maxY += dy*marginScale;
	*maxZ += dz*marginScale;
	
	
}

/*----------------------------------------------------------------------------*/
/** Check that all the points are in the domain [0,X], [0,Y], [0,Z]
 *  Returns 1 if true, 0 if false
 *
 */
int check_points_are_in_domain(double * point, int numberOfPoints, double X, double Y, double Z){
	int i;
	int check=1;
	double x,y,z;

	for(i=0; i<numberOfPoints; i++){
		x=point[i*3]; y=point[i*3+1]; z=point[i*3+2];
		if (x<0 || x>X || y<0 || y>Y || z<0 || z>Z){
			check=0;
			break;
		}
	}

	return check;
}

/*----------------------------------------------------------------------------*/
/** Translate the points by (tX,tY,tZ)
 *
 */
void translate_points(double * point, int numberOfPoints, double tX, double tY, double tZ){
	int i;
	for(i=0; i<numberOfPoints; i++){
		point[i*3]   +=tX;
		point[i*3+1] +=tY;
		point[i*3+2] +=tZ;	
	}
}

/*----------------------------------------------------------------------------*/
/** Translate the points in the alignments results by (tX,tY,tZ)
 *
 */
void translate_points_in_alignments(double * alignments, int numberOfAlignments, double tX, double tY, double tZ){
	int i;
	for(i=0; i<numberOfAlignments; i++){
		alignments[i*10]  +=tX; 
		alignments[i*10+1]+=tY;
		alignments[i*10+2]+=tZ; 
		alignments[i*10+3]+=tX; 
		alignments[i*10+4]+=tY;
		alignments[i*10+5]+=tZ; 
	}
}

/*----------------------------------------------------------------------------*/
/*                                    Main                                    */
/*----------------------------------------------------------------------------*/
int main(int argc, char ** argv)
{
  double * point;
  int N,sectors;
  double Cx,Dx,Cy,Dy,Cz,Dz;
  double X,Y,Z;
  double * align;
  int Nout;
  int indexOfPoint1, indexOfPoint2;

  /* read input */
  if( argc < 9 ) error("\n\nUsage:\n"
		               "point_alignments_3d points.txt Cx Dx Cy Dy Cz Dz output.txt [sectors] [indexOfPoint1 indexOfPoint2]\n\n"
		  	  	  	   "points should be in the domain [Cx Dx] x [Cy Dy] x [Cz Dz]\n"
					   "if no domain is supplied ( Cx Dx Cy Dy Cz Dz == 0 0 0 0 0 0 ), the limits of the data are used \n"
		               "sectors (optional): number of sectors to divide the local cylinder into (from 1 to 360,default=4)\n"
		               "indexOfPoint1 indexOfPoint2 (optional): in order to test a single alignment \n"
		              );
  point = read_txt_handle_comments(argv[1],&N); /* N is the total number of numbers,
                                   	   	   	   	   it corresponds to N/3 points in 3D */
  Cx = atof(argv[2]);
  Dx = atof(argv[3]);
  Cy = atof(argv[4]);
  Dy = atof(argv[5]);
  Cz = atof(argv[6]);
  Dz = atof(argv[7]);

  // if no limits, use the limits of the data
  if (Cx==0 && Dx==0 && Cy==0 && Dy==0 && Cz==0 && Dz==0){
	get_points_domain(point,N/3,&Cx, &Dx, &Cy, &Dy, &Cz, &Dz);
  }

  X=Dx-Cx;
  Y=Dy-Cy;
  Z=Dz-Cz;

  //move the points to the positive domain [0 X] x [0 Y] x [0 Z]
  translate_points(point,N/3,-Cx,-Cy,-Cz); 

  if(!check_points_are_in_domain(point,N/3,X,Y,Z)){
	  error("Points must be in the domain [Cx Dx] x [Cy Dy] x [Cz Dz]\n");
  }

  

  if (argc >= 10){
	  sectors = atoi(argv[9]);
	  if (sectors<1 || sectors>360)
		  sectors = 4;
  }
  else{
	  sectors = 4;
  }

  if (argc == 12){ //in order to test only one alignment
	  indexOfPoint1 = atoi(argv[10]);
	  indexOfPoint2 = atoi(argv[11]);
  }
  else{
	  indexOfPoint1 = -1;
	  indexOfPoint2 = -1;
  }

  printf("\n");
  printf("-------------------------------------------------------------------------------\n");
  printf("Input file : %s\n",argv[1]);
  printf("Output file: %s\n",argv[8]);
  printf("Points: %d\n", N/3);
  printf("Domain [%.5f,%.5f] x [%.5f,%.5f] x [%.5f,%.5f] \n", Cx,Dx,Cy,Dy,Cz,Dz);
  printf("Sectors: %d\n", sectors);

  int testSingleAlignment = indexOfPoint1>0 && indexOfPoint2>0;
  if (testSingleAlignment)
	  printf("Testing single alignment ->  index1: %d, index2: %d \n", indexOfPoint1, indexOfPoint2);
  printf("-------------------------------------------------------------------------------\n");

  /* compute alignments */
  align = point_alignments(point,N/3,X,Y,Z,&Nout, sectors, indexOfPoint1, indexOfPoint2);

  // move the points in the alignments from the positive domain [0 X] x [0 Y] x [0 Z]
  // back to the original domain [Cx Dx] x [Cy Dy] x [Cz Dz]
  translate_points_in_alignments(align,Nout,Cx,Cy,Cz); 

  /* output */
  printf("-------------------------------------------------------------------------------\n");
  printf("%d point alignments found\n",Nout);
  printf("-------------------------------------------------------------------------------\n");
  write_txt(align,10,Nout,argv[8]);


  /* free memory */
  if( align != NULL ) free( (void *) align );
  free( (void *) point );

  return EXIT_SUCCESS;
}
/*----------------------------------------------------------------------------*/
